
{%- for toc_item in toc_item.children %}
    <li class="toctree-l{{ navlevel}}"><a class="reference internal" href="{% if not nav_item == page %}{{ nav_item.url|url }}{% endif %}{{ toc_item.url }}">{{ toc_item.title }}</a>
    {%- set navlevel = navlevel + 1 %}
    {%- if navlevel <= config.theme.navigation_depth and toc_item.children %}
        <ul>
            {%- include 'toc.html' %}
        </ul>
    {%- endif %}
    {%- set navlevel = navlevel - 1 %}
    </li>
{%- endfor %}
