/**
Add macros of the following format to the exported array to define your own macros.
{
    hotkey: 'i',
    name: 'logger.info',
    description: '',
    callback: (cm, event, macroPanel) => {
        // do something
    }
}
*/
export default []
